
from asyncio import Event
from multiprocessing.connection import Connection
import traceback
from typing import Any, Callable, List, Tuple, Optional, AsyncGenerator
import os
import importlib.util
import uuid
import shutil
from appbridge.cipher import Cipher
from appbridge.rsa_key import RSAKey

from reactivestreams.publisher import Publisher
from reactivestreams.subscriber import Subscriber
from reactivestreams.subscription import Subscription
from rsocket.streams.stream_from_async_generator import StreamFromAsyncGenerator
from rsocket.fragment import Fragment
from rsocket.payload import Payload

from appbridge.progress_bar import TranferBar
from appbridge.prototol import AppName, ConnectMsg, DataType, Msg, MsgType, TransData
from appbridge import common
from appbridge.config import Config



#
# client request get file list from me
#
# list <path>
#
async def handle_list_files(msg: Msg, received_files: List[str]) -> AsyncGenerator[Msg, None]:
    files = []
    p = ''
    params = msg.params
    if len(params) > 0:
        p = params[0]
    dir = os.path.join(Config.files_dir, p)

    for item in os.listdir(dir):
        files.append(os.path.join(p, item))

    yield Msg(MsgType.Reply, msg.app, msg.action, files)


#
# otherside request get files/dirs from me
#
# get <file_path_on_otherside>...
#
async def handle_get_files(msg: Msg, received_files: List[str]) -> AsyncGenerator[Msg, None]:
    yield Msg(MsgType.Reply, msg.app, msg.action, files=msg.params)

#
# otherside request send files to me
#
# send <file_path_on_otherside>...
#
async def handle_sendfiles(msg: Msg, received_files: List[str]) -> AsyncGenerator[Msg, None]:

    # save received files to files dir
    common.save_files(received_files)
    
    yield Msg(MsgType.Reply, msg.app, msg.action)


class ChannelSubscriber(Subscriber):

    request_file_path: str = ''
    file_received_size: int = 0
    file_total_size: int = 0
    file = None

    # 正在接收的目录名
    received_dirname:str = ''

    # 正在接收的文件完整路径
    received_file: str = ''

    pending_datatype: DataType
    pending_msg: Any

    session_dir = ''

    finished = False

    cipher: Optional[Cipher]

    file_write_pipe: Connection

    def __init__(self, rsa_key: RSAKey, init_msg, handle_reply, is_service) -> None:
        super().__init__()

        self.rsa_key = rsa_key
        self.is_service = is_service

        self.channel_complete_event = Event()
        self.got_file_event = Event()
        self.got_file_event.set()
        self.request_event = Event()

        self.handle_reply = handle_reply

        self.pending_msg = init_msg

        id = uuid.uuid1()
        session_dir = os.path.join(Config.config_dir, 'sessions', str(id)) 
        if not os.path.isdir(session_dir):
            os.makedirs(session_dir, 0o755)
            print('create session_dir: %s' % session_dir)
        self.session_dir = session_dir
        
        print('[+] Channel subscriber init')


    def __del__(self):

        if self.finished:
            print('This channel already finished!')
            return

        self.finished = True

        self.channel_complete_event.set()

        if self.file:
            self.file.close()

        shutil.rmtree(self.session_dir, ignore_errors=True)
        print('delete session_dir: %s' % self.session_dir)

        print('[-] Channel subscriber deleted')

    def finish(self):
        self.__del__()

    def on_subscribe(self, subscription: Subscription):
        self.subscription = subscription
        subscription.request(1)

    def on_error(self, exception: Exception):
        print('on_error: ' + str(exception))
        # self.pending_msg = Msg.error(str(exception))
        # self.request_event.set()
        self.finish()

    def on_complete(self):
        print('on_complete')
        self.finish()

    def on_pending_msg(self, datatype, msg):
        self.pending_datatype = datatype
        self.pending_msg = msg
        self.request_event.set()

    def on_next(self, value: Payload, is_complete=False):

        # print("From client: len=%d, complete=%s" % (len(value.data), "Yes" if is_complete else "No"))

        if not is_complete:
            self.subscription.request(1)

        # complete signal
        datalen = len(value.data)
        if datalen == 0 and is_complete:
            print("on_next: complete empty msg")
            self.finish()
            return

        # connect/refuse
        data = TransData.fromBytes(value.data)
        # print('transdata1[%d]' % len(data.msg), data.msg)
        if data.type == DataType.connect or data.type == DataType.refuse:
            self.on_pending_msg(data.type, data.resolveMsg()) 
            return

        # file receiving
        if data.type == DataType.encrypted_bin:
            # print('file data msg[%d]' % len(data.msg), data.msg)
            self.file_write_pipe.send(data.msg)
            # bin: bytes = data.resolveMsg(self.file_cipher)
            # if self.file_received_size < self.file_total_size:
            #     self.file.write(bin)
            #     self.file_received_size += datalen
            #     self.bar.next(datalen)

            #     if self.file_received_size >= self.file_total_size:
            #         # received complete
            #         self.file.close()
            #         self.file = None
            #         self.file_received_size = 0
            #         self.file_total_size = 0
            #         self.bar.finish()
            #         print("Received a file!")
            return

        # handle msg
        msg: Msg = data.resolveMsg(self.cipher)
        print('income Msg: %s' % msg.toString())

        if msg.type == MsgType.RequestFile or msg.type == MsgType.Reply or msg.type == MsgType.Request:
            self.on_pending_msg(data.type, msg)

        elif msg.type == MsgType.ReplyFileCipher:
            key = bytes.fromhex(msg.params[0])
            iv = bytes.fromhex(msg.params[1])
            self.file_cipher = Cipher(key, iv)

            # start a new writer process
            (c, process) = common.start_file_writer_process(key, iv)
            self.file_write_process = process
            self.file_write_pipe= c

        elif msg.type == MsgType.ReplyDirInfo:
            dirname = msg.params[0]
            newpath = os.path.join(self.session_dir, dirname)

            index = 0
            while os.path.exists(newpath):
                index += 1
                dirname = os.path.join('__temp_' + str(index), msg.params[0])
                newpath = os.path.join(self.session_dir, dirname)

            if not os.path.isdir(newpath):
                os.makedirs(newpath, 0o755)

            self.received_dirname = dirname  # os.path.basename(newpath)

            print("Prepared a received dir! %s" % self.received_dirname)

        elif msg.type == MsgType.ReplyFileInfo:
            rpath = msg.params[0]
            
            fpath = os.path.join(self.session_dir, self.received_dirname, rpath)

            index = 1
            while os.path.exists(fpath):
                fpath = os.path.join(self.session_dir, self.received_dirname, '__temp_' + str(index), rpath)
                index += 1

            fpath_dir = os.path.dirname(fpath)
            if not os.path.isdir(fpath_dir):
                os.makedirs(fpath_dir, 0o755)

            print("got File info: %s" % fpath)
            self.file_write_pipe.send((fpath, msg.params[1]))

            # self.file_total_size = msg.params[1]
            # self.file =  open(fpath, 'wb')
            self.received_file = fpath

            # # (self.file, self.file_total_size) = common.prepare_receive_file(msg, self.received_path)
            # self.file_received_size = 0
            # self.bar = TranferBar('Receiving', max=self.file_total_size)
            # print("File: %s" % fpath)

        elif msg.type == MsgType.ReplyFileEnd:
            # self.received_dirname = ''
            self.file_write_pipe.send(('',0))
            self.file_write_process.join()
            self.on_pending_msg(data.type, msg)
        
        if is_complete:
            print("got complete msg to finish")
            self.finish()
            print("received a complete request!")
            self.on_pending_msg(data.type, msg)
            # self.pending_msg = msg
            # self.request_event.set()


    def publisher(self) -> Publisher:

        async def request_files(files: List[str], received_files: List[str] = []) -> AsyncGenerator[bytes, None]:
            
            print('Try request files...')
            # await self.got_file_event.wait()

            # print("go on!")
            for f in files:

                # self.got_file_event.clear()
                self.received_file = ''
                # self.received_dir = self.session_dir

                # do request files
                yield TransData.encrypted(Msg(MsgType.RequestFile, params=[f]), self.cipher).toBytes()

                # wait for file received
                print('waiting for file...')
                async for e in wait_for(MsgType.ReplyFileEnd):
                    yield e

                # await self.got_file_event.wait()
                print('GOT file!')
                if self.received_dirname:
                    received_files.append(os.path.join(self.session_dir, self.received_dirname))
                    self.received_dirname = ''
                else:
                    received_files.append(self.received_file)


        async def wait_for_connect():
            while True:
              
                print('WATING for connect reply')
                await self.request_event.wait()
                self.request_event.clear()

                if self.pending_datatype == DataType.connect or self.pending_datatype == DataType.refuse:
                    break

            print('Got connect reply!')


        async def wait_for(type: MsgType, name: str = '') -> AsyncGenerator[bytes, None]:
            while True:
              
                print('WATING for %s' % (type))
                await self.request_event.wait()
                self.request_event.clear()

                if self.pending_datatype != DataType.encrypted:
                    continue

                msg: Msg = self.pending_msg
                # print('got %s' % msg.toString())

                if msg.type == type:
                    break

                # elif msg.type == MsgType.Error:
                #     break

                if msg.type == MsgType.RequestFile:
                    # send file/dir
                    p = msg.params[0]
                    # full_path = os.path.abspath(p)
                    full_path = os.path.join(Config.files_dir, p)
                    norm_path = os.path.abspath(full_path)
                    # basename = os.path.basename(norm_path)
                    # dir = os.path.dirname(norm_path)
                    async for e in common.file_generator(norm_path, self.cipher):
                        yield e
               
            print('GOT %s!' % (msg.type))


        async def handle_request(connect_msg: ConnectMsg) -> AsyncGenerator[bytes, None]:
     
            print('handle request')

            # reply connection
            pubkey = bytes.fromhex(connect_msg.pubkey)
            clientRSA = RSAKey.fromBytes(None, pubkey)

            # random a cipher

            self.cipher = Cipher.new()

            ekey = clientRSA.encryptAsHex(self.cipher.key)
            eiv = clientRSA.encryptAsHex(self.cipher.iv)
            yield TransData.connect(self.rsa_key.pubkey_hex, ekey, eiv).toBytes()

            # waiting for request
            async for e in wait_for(MsgType.Request):
                yield e

            msg: Msg = self.pending_msg
            print('request:', msg.toPrettyString())

            # receive all files from request
            received_files = []
            if msg.files and len(msg.files) > 0:
                # receive all files into this session
                async for e in request_files(msg.files, received_files):
                    yield e

            # find an handler to handle the request
            handle = None
            action = msg.action
            module_file_path = os.path.join(Config.config_dir, action + '.py')
            module_name = action
            if Config.config_dir == '' or not os.path.isfile(module_file_path):
                if Config.files_dir != '':
                    # intergrated action handlers
                    if action == AppName.GetFiles:
                        handle = handle_get_files
                    elif action == AppName.SendFiles:
                        handle = handle_sendfiles
                    elif action == AppName.ListFiles:
                        handle = handle_list_files      
            else:
                # custom action handlers
                print("load %s from %s" % (module_name, module_file_path))
                module_spec = importlib.util.spec_from_file_location(module_name , module_file_path)
                module = importlib.util.module_from_spec(module_spec)
                module_spec.loader.exec_module(module)
                handle = module.handle_action

            if handle is None:
                raise Exception("Don't support app \"%s\"" % action)

            async for m in handle(msg, received_files):
                yield TransData.encrypted(m, self.cipher).toBytes()

            # waiting for channel complete
            async for e in wait_for(MsgType.Complete, msg.app):
                yield e

            print('session end!')

        async def initiate_request(init_msg: Msg) -> AsyncGenerator[bytes, None]:
            
            # waiting for connect/refuse
            await  wait_for_connect()

            if self.pending_datatype == DataType.refuse:
                print('server refused my connection: %s', self.pending_msg.reason)
                return

            msg: ConnectMsg = self.pending_msg
            # got pubkey, key, iv
    
            # print('got connect msg:', msg.toString())
            iv =  self.rsa_key.decrypt(bytes.fromhex(msg.iv))
            key = self.rsa_key.decrypt(bytes.fromhex(msg.key))
            self.cipher = Cipher(key, iv)

            # send request now
            yield  TransData.encrypted(init_msg, self.cipher).toBytes()

            # waiting for reply
            async for e in wait_for(MsgType.Reply):
                yield e

            # handle reply
            msg: Msg = self.pending_msg
            received_files = []
            if msg.files and len(msg.files) > 0:
                # receive all files into this session
                async for e in request_files(msg.files, received_files):
                    yield e

            # custom handle reply
            if self.handle_reply:
                # print('call custom handle reply')
                self.handle_reply(msg, received_files)

        async def generator() -> AsyncGenerator[Tuple[Fragment, bool], None]:
            try: 
                print('generator start...')
                if self.is_service:
                    async for  e in handle_request(self.pending_msg):
                        yield Fragment(e), False
                else:
                    async for  e in initiate_request(self.pending_msg):
                        yield Fragment(e), False
                    yield Fragment(TransData.encrypted(Msg.complete(), self.cipher).toBytes()), True
            except Exception as e:
                traceback.print_exc()
                raise e

        return StreamFromAsyncGenerator(generator)

    
def create(rsa_key: RSAKey, init_msg, handle_reply = None, is_service = False ) -> Tuple[Optional[Publisher], Optional[Subscriber]]:
 
    subscriber = ChannelSubscriber(rsa_key, init_msg, handle_reply, is_service)

    return subscriber.publisher(), subscriber
