

import base64
from ctypes import memmove
from enum import Enum
import json
from typing import AsyncGenerator, Callable, List, Optional


class MsgType(Enum): 
    GetInfo = 'get_info'

    Request = 'request'
    Reply = 'reply'
    Complete = 'complete'

    RequestFile = 'request_file'
    ReplyFileInfo = 'reply_file_info'
    ReplyDirInfo = 'reply_dir_info'
    ReplyFileEnd = 'reply_file_end'

    ReplyFileCipher = 'reply_file_cipher'

    # Error = 'error'

class MsgTypeEncoder(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, MsgType):
      return obj.value
    return json.JSONEncoder.default(self, obj)

class AppName:
    SendFiles = 'send_files'
    GetFiles = 'get_files'
    ListFiles = 'list_files'

class Msg:

    type: MsgType

    app: Optional[str]
    action: Optional[str]
    params: Optional[List[str]]
    files: Optional[List[str]]
  
    def __init__(self, type: MsgType, app: Optional[str] = None,
                        action: Optional[str] = None,
                        params:  Optional[List[str]] = None,
                        files:  Optional[List[str]] = None) -> None:
        self.type = type
        self.app = app
        self.action = action
        self.params = params
        self.files = files


    @classmethod
    def request(cls, app = None, action = None, params = None, files = None):
        return Msg(MsgType.Request, app, action, params, files)

    @classmethod
    def reply(cls, app = None, action = None,  params = None, files = None):
        return Msg(MsgType.Reply, app, action, params, files)

    @classmethod
    def complete(cls):
        return Msg(MsgType.Complete)

    # @classmethod
    # def error(cls, msg: str):
    #     return Msg(MsgType.Error, params=[msg])

    @classmethod
    def fromBytes(cls, data: bytes):
        try:
            string = data.decode(encoding='UTF-8', errors='strict')
            obj = json.loads(string)

            # t = obj['type']
            # print('type=%s' % t)
            # tt = MsgType[t]

            return Msg( MsgType(obj['type']), obj['app'],  obj['action'], obj['params'], obj['files'])
        except Exception as e:
            print(e)
            raise Exception("data format error")


    def toDict(self) -> dict:
        return {
            'type': self.type,
            'app': self.app,
            'action': self.action,
            'params': self.params,
            'files': self.files,
        }

    def toString(self) -> str:
        return json.dumps(self.toDict(), cls=MsgTypeEncoder)

    def toPrettyString(self) -> str:
        return json.dumps(self.toDict(), cls=MsgTypeEncoder, indent='    ')

    def toBytes(self) -> bytes:
        return self.toString().encode(encoding='UTF-8', errors='strict')



class DataType(Enum): 
    connect = 0
    refuse = 1
    encrypted = 2
    encrypted_bin = 3

class TransData:

    type: DataType
    msg: bytes
    
    def __init__(self, type: DataType, msg: bytes) -> None:
        self.type = type
        self.msg = msg
        # mv = memoryview(msg)
        # if end > 0:
        #     self.msg = mv[start:end]
        # else:
        #     if start > 0:
        #         self.msg = mv[start:]
        #     else:
        #         self.msg = mv

    @classmethod
    def connect(cls, pubkey_hex: str, 
            key: Optional[str] = None, iv: Optional[str] = None):
        msg = ConnectMsg(pubkey_hex, key, iv)
        return TransData(DataType.connect, msg.toBytes())
    
    @classmethod
    def refuse(cls, reason):
        msg = {
            reason: reason,
        }
        return TransData(DataType.refuse, 
                    bytes(json.dumps(msg), encoding='UTF-8'))
      
    @classmethod
    def encrypted(cls, msg: Msg, cipher):
        return TransData(DataType.encrypted, 
                    cipher.encrypt(msg.toBytes()))

    @classmethod
    def encrypted_bin(cls, bin: bytes, cipher):
        return TransData(DataType.encrypted_bin, 
                    cipher.encrypt(bin))  

    @classmethod
    def fromBytes(cls, data: bytes):
        # print("transdata[%d]" % len(data), data)
        try:
            data_type = int.from_bytes(data[:4], 'little', signed=False)
            # print('data type = ', data_type)
            return TransData(DataType(data_type), data[4:])
        except Exception as e:
            print(e)
            print('data: ', data)
            raise Exception("data format error")

    def toBytes(self) -> bytes:
        data_type: int = self.type.value
        return data_type.to_bytes(4, 'little', signed=False) + self.msg

    def resolveMsg(self, cipher = None):
        # print('data type = ', self.type.value)
        if self.type == DataType.connect:
            return ConnectMsg.fromBytes(self.msg)
        if self.type == DataType.refuse:
            return RefuseMsg.fromBytes(self.msg)
    
        decrypted_msg: bytes = cipher.decrypt(self.msg)
        if self.type == DataType.encrypted_bin:
            return decrypted_msg

        return Msg.fromBytes(decrypted_msg)


class ConnectMsg:
    pubkey: str # pubkey.pem hex
    key: Optional[str]
    iv: Optional[str]
  
    def __init__(self, pubkey: str, 
                key: Optional[str] = None, 
                iv: Optional[str] = None) -> None:
        self.pubkey = pubkey
        self.key = key
        self.iv = iv

    @classmethod
    def fromBytes(cls, data: bytes):
        # print("connect msg from bytes...")
        try:
            string = data.decode(encoding='utf8', errors='strict')
            obj = json.loads(string)
            return ConnectMsg(obj['pubkey'], 
                                    obj['aes']['key'],
                                    obj['aes']['iv'])
        except Exception as e:
            print(e)
            raise Exception("data format error")


    def toDict(self) -> dict:
        return {
            'pubkey': self.pubkey,
            'aes': {
                'key': self.key,
                'iv': self.iv,
            }
        }

    def toString(self) -> str:
        return json.dumps(self.toDict(), cls=MsgTypeEncoder)

    def toPrettyString(self) -> str:
        return json.dumps(self.toDict(), cls=MsgTypeEncoder, indent='    ')

    def toBytes(self) -> bytes:
        return self.toString().encode(encoding='utf8', errors='strict')



class RefuseMsg:
    reason: str
  
    def __init__(self, reason: str) -> None:
        self.reason = reason

    @classmethod
    def fromBytes(cls, data: bytes):
        try:
            string = data.decode(encoding='utf8', errors='strict')
            obj = json.loads(string)
            return RefuseMsg( MsgType(obj['reason']))
        except Exception as e:
            print(e)
            raise Exception("data format error")


    def toDict(self) -> dict:
        return {
            'reason': self.reason,
        }

    def toString(self) -> str:
        return json.dumps(self.toDict(), cls=MsgTypeEncoder)

    def toPrettyString(self) -> str:
        return json.dumps(self.toDict(), cls=MsgTypeEncoder, indent='    ')

    def toBytes(self) -> bytes:
        return self.toString().encode(encoding='utf8', errors='strict')


HandleReplyCallable = Optional[Callable[[Msg, List[str]], None]]
